from decimal import Decimal
from testfixtures import RoundComparison as R, compare, ShouldRaise
from unittest import TestCase

from ..compat import PY2, PY3


class Tests(TestCase):

    def test_equal_yes_rhs(self):
        self.assertTrue(0.123457 == R(0.123456, 5))

    def test_equal_yes_lhs(self):
        self.assertTrue(R(0.123456, 5) == 0.123457)

    def test_equal_no_rhs(self):
        self.assertFalse(0.123453 == R(0.123456, 5))

    def test_equal_no_lhs(self):
        self.assertFalse(R(0.123456, 5) == 0.123453)

    def test_not_equal_yes_rhs(self):
        self.assertFalse(0.123457 != R(0.123456, 5))

    def test_not_equal_yes_lhs(self):
        self.assertFalse(R(0.123456, 5) != 0.123457)

    def test_not_equal_no_rhs(self):
        self.assertTrue(0.123453 != R(0.123456, 5))

    def test_not_equal_no_lhs(self):
        self.assertTrue(R(0.123456, 5) != 0.123453)

    def test_equal_in_sequence_rhs(self):
        self.assertEqual((1, 2, 0.123457),
                         (1, 2, R(0.123456, 5)))

    def test_equal_in_sequence_lhs(self):
        self.assertEqual((1, 2, R(0.123456, 5)),
                         (1, 2, 0.123457))

    def test_not_equal_in_sequence_rhs(self):
        self.assertNotEqual((1, 2, 0.1236),
                            (1, 2, R(0.123456, 5)))

    def test_not_equal_in_sequence_lhs(self):
        self.assertNotEqual((1, 2, R(0.123456, 5)),
                            (1, 2, 0.1236))

    def test_not_numeric_rhs(self):
        with ShouldRaise(TypeError):
            'abc' == R(0.123456, 5)

    def test_not_numeric_lhs(self):
        with ShouldRaise(TypeError):
            R(0.123456, 5) == 'abc'

    def test_repr(self):
        compare('<R:0.12346 to 5 digits>',
                repr(R(0.123456, 5)))

    def test_str(self):
        compare('<R:0.12346 to 5 digits>',
                repr(R(0.123456, 5)))

    def test_str_negative(self):
        if PY3:
            expected = '<R:123500 to -2 digits>'
        else:
            expected = '<R:123500.0 to -2 digits>'
        compare(expected, repr(R(123456, -2)))

    TYPE_ERROR_DECIMAL = TypeError(
        "Cannot compare <R:0.12346 to 5 digits> with <class 'decimal.Decimal'>"
        )

    def test_equal_yes_decimal_to_float_rhs(self):
        with ShouldRaise(self.TYPE_ERROR_DECIMAL, unless=PY2):
            self.assertTrue(Decimal("0.123457") == R(0.123456, 5))

    def test_equal_yes_decimal_to_float_lhs(self):
        with ShouldRaise(self.TYPE_ERROR_DECIMAL, unless=PY2):
            self.assertTrue(R(0.123456, 5) == Decimal("0.123457"))

    def test_equal_no_decimal_to_float_rhs(self):
        with ShouldRaise(self.TYPE_ERROR_DECIMAL, unless=PY2):
            self.assertFalse(Decimal("0.123453") == R(0.123456, 5))

    def test_equal_no_decimal_to_float_lhs(self):
        with ShouldRaise(self.TYPE_ERROR_DECIMAL, unless=PY2):
            self.assertFalse(R(0.123456, 5) == Decimal("0.123453"))

    TYPE_ERROR_FLOAT = TypeError(
        "Cannot compare <R:0.12346 to 5 digits> with <class 'float'>"
        )

    def test_equal_yes_float_to_decimal_rhs(self):
        with ShouldRaise(self.TYPE_ERROR_FLOAT, unless=PY2):
            self.assertTrue(0.123457 == R(Decimal("0.123456"), 5))

    def test_equal_yes_float_to_decimal_lhs(self):
        with ShouldRaise(self.TYPE_ERROR_FLOAT, unless=PY2):
            self.assertTrue(R(Decimal("0.123456"), 5) == 0.123457)

    def test_equal_no_float_to_decimal_rhs(self):
        with ShouldRaise(self.TYPE_ERROR_FLOAT, unless=PY2):
            self.assertFalse(0.123453 == R(Decimal("0.123456"), 5))

    def test_equal_no_float_to_decimal_lhs(self):
        with ShouldRaise(self.TYPE_ERROR_FLOAT, unless=PY2):
            self.assertFalse(R(Decimal("0.123456"), 5) == 0.123453)

    def test_integer_float(self):
        with ShouldRaise(TypeError, unless=PY2):
            1 == R(1.000001, 5)

    def test_float_integer(self):
        with ShouldRaise(TypeError, unless=PY2):
            R(1.000001, 5) == 1

    def test_equal_yes_integer_other_rhs(self):
        self.assertTrue(10 == R(11, -1))

    def test_equal_yes_integer_lhs(self):
        self.assertTrue(R(11, -1) == 10)

    def test_equal_no_integer_rhs(self):
        self.assertFalse(10 == R(16, -1))

    def test_equal_no_integer_lhs(self):
        self.assertFalse(R(16, -1) == 10)

    def test_equal_integer_zero_precision(self):
        self.assertTrue(1 == R(1, 0))

    def test_equal_yes_negative_precision(self):
        self.assertTrue(149.123 == R(101.123, -2))

    def test_equal_no_negative_precision(self):
        self.assertFalse(149.123 == R(150.001, -2))

    def test_decimal_yes_rhs(self):
        self.assertTrue(Decimal('0.123457') == R(Decimal('0.123456'), 5))

    def test_decimal_yes_lhs(self):
        self.assertTrue(R(Decimal('0.123456'), 5) == Decimal('0.123457'))

    def test_decimal_no_rhs(self):
        self.assertFalse(Decimal('0.123453') == R(Decimal('0.123456'), 5))

    def test_decimal_no_lhs(self):
        self.assertFalse(R(Decimal('0.123456'), 5) == Decimal('0.123453'))



from django.core.exceptions import ValidationError

from testfixtures import ShouldRaise
from testfixtures.compat import PY2
from testfixtures.shouldraise import ShouldAssert


class TestShouldRaiseWithValidatorErrors(object):

    def test_as_expected(self):
        with ShouldRaise(ValidationError("d'oh")):
            raise ValidationError("d'oh")

    def test_not_as_expected(self):
        if PY2:
            message = (
                'ValidationError([u"d\'oh"]) (expected) != '
                'ValidationError([u\'nuts\']) (raised)'
            )
        else:
            message = (
                'ValidationError(["d\'oh"]) (expected) != '
                'ValidationError([\'nuts\']) (raised)'
            )
        with ShouldAssert(message):
            with ShouldRaise(ValidationError("d'oh")):
                raise ValidationError("nuts")



from unittest import TestCase

import warnings

from testfixtures import (
    ShouldWarn, compare, ShouldRaise, ShouldNotWarn,
    Comparison as C
)
from testfixtures.compat import PY3, PY_36_PLUS
from testfixtures.shouldraise import ShouldAssert

if PY3:
    warn_module = 'builtins'
else:
    warn_module = 'exceptions'


class ShouldWarnTests(TestCase):

    def test_warn_expected(self):
        with warnings.catch_warnings(record=True) as backstop:
            with ShouldWarn(UserWarning('foo')):
                warnings.warn('foo')
        compare(len(backstop), expected=0)

    def test_warn_not_expected(self):
        with ShouldAssert(
            "sequence not as expected:\n\n"
            "same:\n[]\n\n"
            "expected:\n[]\n\n"
            "actual:\n[UserWarning('foo',)]"
        ):
            with warnings.catch_warnings(record=True) as backstop:
                with ShouldNotWarn():
                    warnings.warn('foo')
        compare(len(backstop), expected=0)

    def test_no_warn_expected(self):
        with ShouldNotWarn():
            pass

    def test_no_warn_not_expected(self):
        with ShouldAssert(
            "sequence not as expected:\n\n"
            "same:\n[]\n\n"
            "expected:\n[<C:"+warn_module+".UserWarning>args: ('foo',)</C>]"
            "\n\nactual:\n[]"
        ):
            with ShouldWarn(UserWarning('foo')):
                pass

    def test_filters_removed(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with ShouldWarn(UserWarning("foo")):
                warnings.warn('foo')

    def test_multiple_warnings(self):
        with ShouldRaise(AssertionError) as s:
            with ShouldWarn(UserWarning('foo')):
                warnings.warn('foo')
                warnings.warn('bar')
        content = str(s.raised)
        self.assertTrue('foo' in content)
        self.assertTrue('bar' in content)

    def test_minimal_ok(self):
        with ShouldWarn(UserWarning):
            warnings.warn('foo')

    def test_minimal_bad(self):
        with ShouldAssert(
            "sequence not as expected:\n\n"
            "same:\n[]\n\n"
            "expected:\n"
            "[<C(failed):"+warn_module+".DeprecationWarning>wrong type</C>]\n\n"
            "actual:\n[UserWarning('foo',)]"
        ):
            with ShouldWarn(DeprecationWarning):
                warnings.warn('foo')

    def test_maximal_ok(self):
        with ShouldWarn(DeprecationWarning('foo')):
            warnings.warn_explicit(
                'foo', DeprecationWarning, 'bar.py', 42, 'bar_module'
            )

    def test_maximal_bad(self):
        with ShouldAssert(
            "sequence not as expected:\n\n"
            "same:\n[]\n\n"
            "expected:\n[\n"
            "<C(failed):"+warn_module+".DeprecationWarning>\n"
            "attributes differ:\n"
            "'args': ('bar',) (Comparison) != ('foo',) (actual)\n</C>]\n\n"
            "actual:\n[DeprecationWarning('foo',)]"
        ):
            with ShouldWarn(DeprecationWarning('bar')):
                warnings.warn_explicit(
                    'foo', DeprecationWarning, 'bar.py', 42, 'bar_module'
                )

    def test_maximal_explore(self):
        with ShouldWarn() as recorded:
            warnings.warn_explicit(
                'foo', DeprecationWarning, 'bar.py', 42, 'bar_module'
            )
        compare(len(recorded), expected=1)

        expected_attrs = dict(
            _category_name='DeprecationWarning',
            category=DeprecationWarning,
            file=None,
            filename='bar.py',
            line=None,
            lineno=42,
            message=C(DeprecationWarning('foo')),
        )

        if PY_36_PLUS:
            expected_attrs['source'] = None

        compare(expected=C(warnings.WarningMessage, **expected_attrs),
            actual=recorded[0])

    def test_filter_present(self):
        with ShouldWarn(DeprecationWarning,
                        message="This function is deprecated."):
            warnings.warn("This utility is deprecated.", DeprecationWarning)
            warnings.warn("This function is deprecated.", DeprecationWarning)

    def test_filter_missing(self):
        if PY3:
            type_repr = 'builtins.DeprecationWarning'
        else:
            type_repr = 'exceptions.DeprecationWarning'
        with ShouldAssert(
            "sequence not as expected:\n\n"
            "same:\n[]\n\n"
            "expected:\n[<C:{}>]\n\n"
            "actual:\n[]".format(type_repr)
        ):
            with ShouldWarn(DeprecationWarning,
                            message="This function is deprecated."):
                warnings.warn("This utility is deprecated.", DeprecationWarning)



from calendar import timegm
from datetime import datetime, timedelta, date
from testfixtures.compat import new_class


@classmethod
def add(cls, *args, **kw):
    if 'tzinfo' in kw or len(args) > 7:
        raise TypeError('Cannot add using tzinfo on %s' % cls.__name__)
    if args and isinstance(args[0], cls.__bases__[0]):
        inst = args[0]
        tzinfo = getattr(inst, 'tzinfo', None)
        if tzinfo:
            if tzinfo != cls._tzta:
                raise ValueError(
                    'Cannot add %s with tzinfo of %s as configured to use %s' % (
                        inst.__class__.__name__, tzinfo, cls._tzta
                    ))
            inst = inst.replace(tzinfo=None)
        if cls._ct:
            inst = cls._ct(inst)
        cls._q.append(inst)
    else:
        cls._q.append(cls(*args, **kw))


@classmethod
def set_(cls, *args, **kw):
    if cls._q:
        cls._q = []
    cls.add(*args, **kw)


@classmethod
def tick(cls, *args, **kw):
    if kw:
        delta = timedelta(**kw)
    else:
        delta, = args
    cls._q[-1] += delta


def __add__(self, other):
    r = super(self.__class__, self).__add__(other)
    if self._ct:
        r = self._ct(r)
    return r


def __new__(cls, *args, **kw):
    if cls is cls._cls:
        return super(cls, cls).__new__(cls, *args, **kw)
    else:
        return cls._cls(*args, **kw)


@classmethod
def instantiate(cls):
    r = cls._q.pop(0)
    if not cls._q:
        cls._gap += cls._gap_d
        n = r + timedelta(**{cls._gap_t: cls._gap})
        if cls._ct:
            n = cls._ct(n)
        cls._q.append(n)
    return r


@classmethod
def now(cls, tz=None):
    r = cls._instantiate()
    if tz is not None:
        if cls._tzta:
            r = r - cls._tzta.utcoffset(r)
        r = tz.fromutc(r.replace(tzinfo=tz))
    return cls._ct(r)


@classmethod
def utcnow(cls):
    r = cls._instantiate()
    if cls._tzta is not None:
        r = r - cls._tzta.utcoffset(r)
    return r


def test_factory(n, type, default, args, kw, tz=None, **to_patch):
    q = []
    to_patch['_q'] = q
    to_patch['_tzta'] = tz
    to_patch['add'] = add
    to_patch['set'] = set_
    to_patch['tick'] = tick
    to_patch['__add__'] = __add__
    if '__new__' not in to_patch:
        to_patch['__new__'] = __new__
    class_ = new_class(n, (type, ), to_patch)
    strict = kw.pop('strict', False)
    if strict:
        class_._cls = class_
    else:
        class_._cls = type
    if args != (None, ):
        if not (args or kw):
            args = default
        class_.add(*args, **kw)
    return class_


def correct_date_method(self):
    return self._date_type(
        self.year,
        self.month,
        self.day
        )


@classmethod
def correct_datetime(cls, dt):
    return cls._cls(
        dt.year,
        dt.month,
        dt.day,
        dt.hour,
        dt.minute,
        dt.second,
        dt.microsecond,
        dt.tzinfo,
        )


def test_datetime(*args, **kw):
    if len(args) > 7:
        tz = args[7]
        args = args[:7]
    else:
        tz = kw.pop('tzinfo', getattr(args[0], 'tzinfo', None) if args else None)
    if 'delta' in kw:
        gap = kw.pop('delta')
        gap_delta = 0
    else:
        gap = 0
        gap_delta = 10
    delta_type = kw.pop('delta_type', 'seconds')
    date_type = kw.pop('date_type', date)
    return test_factory(
        'tdatetime', datetime, (2001, 1, 1, 0, 0, 0), args, kw, tz,
        _ct=correct_datetime,
        _instantiate=instantiate,
        now=now,
        utcnow=utcnow,
        _gap=gap,
        _gap_d=gap_delta,
        _gap_t=delta_type,
        date=correct_date_method,
        _date_type=date_type,
        )

test_datetime.__test__ = False


@classmethod
def correct_date(cls, d):
    return cls._cls(
        d.year,
        d.month,
        d.day,
        )


def test_date(*args, **kw):
    if 'delta' in kw:
        gap = kw.pop('delta')
        gap_delta = 0
    else:
        gap = 0
        gap_delta = 1
    delta_type = kw.pop('delta_type', 'days')
    return test_factory(
        'tdate', date, (2001, 1, 1), args, kw,
        _ct=correct_date,
        today=instantiate,
        _gap=gap,
        _gap_d=gap_delta,
        _gap_t=delta_type,
        )

ms = 10**6


def __time_new__(cls, *args, **kw):
    if args or kw:
        return super(cls, cls).__new__(cls, *args, **kw)
    else:
        val = cls.instantiate()
        t = timegm(val.utctimetuple())
        t += (float(val.microsecond)/ms)
        return t

test_date.__test__ = False


def test_time(*args, **kw):
    if 'tzinfo' in kw or len(args) > 7 or (args and getattr(args[0], 'tzinfo', None)):
        raise TypeError("You don't want to use tzinfo with test_time")
    if 'delta' in kw:
        gap = kw.pop('delta')
        gap_delta = 0
    else:
        gap = 0
        gap_delta = 1
    delta_type = kw.pop('delta_type', 'seconds')
    return test_factory(
        'ttime', datetime, (2001, 1, 1, 0, 0, 0), args, kw,
        _ct=None,
        instantiate=instantiate,
        _gap=gap,
        _gap_d=gap_delta,
        _gap_t=delta_type,
        __new__=__time_new__,
        )

test_time.__test__ = False



import os

from testfixtures.shouldraise import ShouldAssert
from .mock import Mock
from tempfile import mkdtemp
from testfixtures import Replacer, ShouldRaise, TempDirectory, compare, tempdir
from unittest import TestCase

from ..rmtree import rmtree


class TestTempDir(TestCase):

    @tempdir()
    def test_simple(self, d):
        d.write('something', b'stuff')
        d.write('.svn', b'stuff')
        d.compare((
            '.svn',
            'something',
            ))

    @tempdir()
    def test_subdirs(self, d):
        subdir = ['some', 'thing']
        d.write(subdir+['something'], b'stuff')
        d.write(subdir+['.svn'], b'stuff')
        d.compare(path=subdir, expected=(
            '.svn',
            'something',
            ))

    @tempdir()
    def test_not_same(self, d):
        d.write('something', b'stuff')

        with ShouldAssert(
            "sequence not as expected:\n"
            "\n"
            "same:\n"
            "()\n"
            "\n"
            "expected:\n"
            "('.svn', 'something')\n"
            "\n"
            "actual:\n"
            "('something',)"
        ):
            d.compare(['.svn', 'something'])

    @tempdir(ignore=('.svn', ))
    def test_ignore(self, d):
        d.write('something', b'stuff')
        d.write('.svn', b'stuff')
        d.compare(['something'])

    def test_cleanup_properly(self):
        r = Replacer()
        try:
            m = Mock()
            d = mkdtemp()
            m.return_value = d
            r.replace('testfixtures.tempdirectory.mkdtemp', m)

            self.failUnless(os.path.exists(d))

            self.assertFalse(m.called)

            @tempdir()
            def test_method(d):
                d.write('something', b'stuff')
                d.compare(['something'])

            self.assertFalse(m.called)
            compare(os.listdir(d), [])

            test_method()

            self.assertTrue(m.called)
            self.failIf(os.path.exists(d))

        finally:
            r.restore()
            if os.path.exists(d):
                # only runs if the test fails!
                rmtree(d)  # pragma: no cover

    @tempdir()
    def test_cleanup_test_okay_with_deleted_dir(self, d):
        rmtree(d.path)

    @tempdir()
    def test_decorator_returns_tempdirectory(self, d):
        # check for what we get, so we only have to write
        # tests in test_tempdirectory.py
        self.failUnless(isinstance(d, TempDirectory))

    def test_dont_create_or_cleanup_with_path(self):
        with Replacer() as r:
            m = Mock()
            r.replace('testfixtures.tempdirectory.mkdtemp', m)
            r.replace('testfixtures.tempdirectory.rmtree', m)

            @tempdir(path='foo')
            def test_method(d):
                compare(d.path, 'foo')

            test_method()

            self.assertFalse(m.called)
